/*	File:	sound.cpp
	Author:	Alex Fedosov
	Date:	October 9, 2000
	Description:	DMZ sound implementation: WAV-file related functions
*/

#include <windows.h>
#include <dsound.h>
#include "dmz_snd.h"
#include "dmz_snd_internal.h"

LPDIRECTSOUNDBUFFER lpsounds[MAXSOUNDS];

int dmz_snd_load(char* filename)
{
	LPVOID write1, write2;
	DWORD length1, length2;
	HRESULT hr;
	HMMIO hmfr;
	MMCKINFO parent, child;
	WAVEFORMATEX wfmtx;
	BYTE *wavdata;
	DSBUFFERDESC dsbd;
	unsigned int i, nxtsnd, size = 0;

	if (lpds == NULL) return DMZ_SND_NOT_INIT;

	// look for first available slot in the array */
	for (i = 1; i < MAXSOUNDS; i++)
		if (lpsounds[i] == NULL) break;

	nxtsnd = i;
	if (nxtsnd >= MAXSOUNDS) return DMZ_SND_TOO_MANY;
		
	parent.ckid = (FOURCC) 0;
	parent.cksize = 0;
	parent.fccType = (FOURCC) 0;
	parent.dwDataOffset = 0;
	parent.dwFlags = 0;

	child = parent;

	// open the WAV file for reading
	hmfr = mmioOpen(filename, NULL, MMIO_READ | MMIO_ALLOCBUF);

	if (hmfr == NULL)
		return DMZ_SND_OPEN_ERROR;			// unable to open file

	// descend into the RIFF
	parent.fccType = mmioFOURCC('W', 'A', 'V', 'E');

	if (mmioDescend(hmfr, &parent, NULL, MMIO_FINDRIFF))
	{
		mmioClose(hmfr, 0);
		return DMZ_SND_BAD_WAVE;			// not a WAV file
	}

	// descend to wave fmt
	child.ckid = mmioFOURCC('f', 'm', 't', ' ');
	if (mmioDescend(hmfr, &child, &parent, 0))
	{
		mmioClose(hmfr, 0);
		return DMZ_SND_BAD_WAVE;			// no fmt chunk
	}

	// read the WAVEFMT from the wave file
	if (mmioRead(hmfr, (char *) &wfmtx, sizeof(wfmtx)) != sizeof(wfmtx))
	{
		mmioClose(hmfr, 0);
		return DMZ_SND_READ_ERROR;			// cannot read from WAV file
	}

	// only handle PCM format WAVs
	if (wfmtx.wFormatTag != WAVE_FORMAT_PCM)
	{
		mmioClose(hmfr, 0);
		return DMZ_SND_BAD_WAVE;			// not PCM format
	}

	// ascend out of the format chunk
	if (mmioAscend(hmfr, &child, 0))
	{
		mmioClose(hmfr, 0);
		return DMZ_SND_BAD_WAVE;			// unable to ascend
	}

	// descend to the data chunk
	child.ckid = mmioFOURCC('d', 'a', 't', 'a');

	if (mmioDescend(hmfr, &child, &parent, MMIO_FINDCHUNK))
	{
		mmioClose(hmfr, 0);
		return DMZ_SND_BAD_WAVE;			// no data chunk
	}

	size = child.cksize;
	wavdata = (BYTE *) malloc(size);
	
	if (wavdata == NULL)
	{
		mmioClose(hmfr, 0);
		return DMZ_SND_NO_MEM;			// no memory available
	}

	// read the WAV data
	if ((DWORD) mmioRead(hmfr, (char *) wavdata, size) != size)
	{
		mmioClose(hmfr, 0);
		return DMZ_SND_READ_ERROR;			// unable to read data chunk
	}

	mmioClose(hmfr, 0);

	// set up the DirectSound buffer description
	memset(&dsbd, 0, sizeof(DSBUFFERDESC));
	dsbd.dwSize = sizeof(DSBUFFERDESC);
	dsbd.dwFlags = DSBCAPS_CTRLDEFAULT | DSBCAPS_STATIC | DSBCAPS_LOCSOFTWARE;
	dsbd.lpwfxFormat = &wfmtx;
	dsbd.dwBufferBytes = size;

	// setup the DirectSound buffer
	if (lpds->CreateSoundBuffer(&dsbd, &lpsounds[nxtsnd], NULL) != DS_OK)
		return DMZ_SND_DS_ERROR;

	// copy the WAV data to the buffer
	hr = lpsounds[nxtsnd]->Lock(0, size, &write1, &length1, &write2, &length2, 0);

	if (hr == DSERR_BUFFERLOST)
	{
		lpsounds[nxtsnd]->Restore();
		hr = lpsounds[nxtsnd]->Lock(0, size, &write1, &length1, &write2, &length2, 0);
	}

	if (hr != DS_OK)
		return DMZ_SND_DS_ERROR;

	CopyMemory(write1, wavdata, length1);

	if (write2 != NULL)
	{
		CopyMemory(write2, wavdata + length1, length2);
	}

	hr = lpsounds[nxtsnd]->Unlock(write1, length1, write2, length2);

	if (hr != DS_OK)
	{
		return DMZ_SND_DS_ERROR;
	}

	return nxtsnd++;
}

int dmz_snd_play(int sid, int flags)
{
	if (lpds == NULL) return DMZ_SND_NOT_INIT;
	if (sid < 0 || sid >= MAXSOUNDS) return DMZ_SND_INVALID_SID;
	if (lpsounds[sid] == NULL) return DMZ_SND_UNUSED_SID;

	flags == DMZ_SND_LOOP ? flags = DSBPLAY_LOOPING : flags = 0;

	lpsounds[sid]->Stop();

	return (lpsounds[sid]->Play(0, 0, flags) == DS_OK) ?
			DMZ_SND_SUCCESS : DMZ_SND_DS_ERROR;
}

int dmz_snd_stop(int sid)
{
	if (lpds == NULL) return DMZ_SND_NOT_INIT;
	if (sid < 0 || sid >= MAXSOUNDS) return DMZ_SND_INVALID_SID;
	if (lpsounds[sid] == NULL) return DMZ_SND_UNUSED_SID;

	return (lpsounds[sid]->Stop() == DS_OK) ?
			DMZ_SND_SUCCESS : DMZ_SND_DS_ERROR;
}

int dmz_snd_unload(int sid)
{
	if (lpds == NULL) return DMZ_SND_NOT_INIT;
	if (sid < 0 || sid >= MAXSOUNDS) return DMZ_SND_INVALID_SID;
	if (lpsounds[sid] == NULL) return DMZ_SND_UNUSED_SID;

	lpsounds[sid]->Stop();
	
	DXSAFERELEASE(lpsounds[sid]);
	current_volume[sid] = 0;

	return DMZ_SND_SUCCESS;
}
