/*	File:	bitmap.cpp
	Author:	Alex Fedosov
	Date:	March 14, 2001
	Description:	code for loading bitmaps and creating surfaces
*/

#include <windows.h>
#include <ddraw.h>

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "dmzgraph.h"
#include "dmzgraph_internal.h"

int numRedBits, numGrnBits, numBluBits;
int loRedBit, loGrnBit, loBluBit;

int dmz_graph_load_image(char *filename, int *width, int *height)
{
	BYTE *surf;
	DDSURFACEDESC desc;
	HRESULT r;
	int w, h, err;
	int imgID = get_available_image();
	unsigned char *bitmap;

	if (!init_done) return DMZ_GRAPH_NOT_INIT;
	if (imgID < 0) return DMZ_GRAPH_TOO_MANY;

	bitmap = dmz_graph_load_bitmap(filename, &w, &h);
	if (bitmap == NULL)
		return DMZ_GRAPH_LOAD_ERROR;

	/* hack to allow loading without using DDraw */
	if (display_depth == 0)
	{
		if (width != NULL) *width = w;
		if (height != NULL) *height = h;
		images[imgID] = (LPDIRECTDRAWSURFACE) ~ (int) images[imgID];
		free(bitmap);
		return imgID;
	}

	ZeroMemory(&desc, sizeof(desc));
	desc.dwSize = sizeof(desc);
	desc.dwFlags = DDSD_WIDTH | DDSD_HEIGHT | DDSD_CAPS;
	desc.dwWidth = w;
	desc.dwHeight = h;
	desc.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN | DDSCAPS_SYSTEMMEMORY;

	r = ddraw->CreateSurface(&desc, &images[imgID], 0);

	if (r != DD_OK)
	{
		last_error = 0 - (r & 0xFFFF);
		return DMZ_GRAPH_DDRAW_ERROR;
	}

	r = images[imgID]->Lock(NULL, &desc, DDLOCK_WAIT | DDLOCK_WRITEONLY, 0);
	if (r != DD_OK)
	{
		last_error = 0 - (r & 0xFFFF);
		return DMZ_GRAPH_DDRAW_ERROR;
	}

	surf = (BYTE *) desc.lpSurface;

	switch (display_depth)
	{
	case 16:
		__img24_srf16(bitmap, surf, w, h, desc.lPitch);
		err = imgID;
		break;
	case 24:
		__img24_srf24(bitmap, surf, w, h, desc.lPitch);
		err = imgID;
		break;
	case 32:
		__img24_srf32(bitmap, surf, w, h, desc.lPitch);
		err = imgID;
		break;
	default:
		err = DMZ_GRAPH_UNSUPPORTED;
		break;
	}

	images[imgID]->Unlock(0);

	if (width != NULL) *width = w;
	if (height != NULL)
	{
		if (h < 0) *height = 0 - h;
		else *height = h;
	}

	free(bitmap);
	return err;
}

inline void __img24_srf16(unsigned char* bitmap, BYTE *surf, 
						int width, int height, LONG lPitch)
{
	float red_div = (float) 256 / (float) pow(2, numRedBits);
	float grn_div = (float) 256 / (float) pow(2, numGrnBits);
	float blu_div = (float) 256 / (float) pow(2, numBluBits);

	for (int i = height - 1; i >= 0; i--)
	{
		RGBTRIPLE *rgb = (RGBTRIPLE *) (bitmap + i * width * 3);
		USHORT *lpwSurf = (USHORT *) surf;

		for (int j = 0; j < width; j++, rgb++)
		{
			float rf = (float) rgb->rgbtRed / red_div;
			float gf = (float) rgb->rgbtGreen / grn_div;
			float bf = (float) rgb->rgbtBlue / blu_div;

			WORD r = (WORD) ((WORD) rf << loRedBit);
			WORD g = (WORD) ((WORD) gf << loGrnBit);
			WORD b = (WORD) ((WORD) bf << loBluBit);

			lpwSurf[j] = (WORD) (r | g | b);
		}
		surf += lPitch;
	}
}

inline void __img24_srf24(unsigned char* bitmap, BYTE *surf, 
						  int width, int height, LONG lPitch)
{
	for (int i = height - 1; i >= 0; i--)
	{
		RGBTRIPLE *rgb = (RGBTRIPLE *) (bitmap + i * width * 3);
		RGBTRIPLE *dst = (RGBTRIPLE *) surf;

		for (int j = 0; j < width; j++, rgb++)
		{
			DWORD r = rgb->rgbtRed << loRedBit;
			DWORD g = rgb->rgbtGreen << loGrnBit;
			DWORD b = rgb->rgbtBlue << loBluBit;
			DWORD *lpdwSurf = (DWORD *) dst;

			*lpdwSurf = r | g | b;
			dst++;
		}
		surf += lPitch;
	}
}

inline void __img24_srf32(unsigned char* bitmap, BYTE *surf, 
						int width, int height, LONG lPitch)
{
	for (int i = height - 1; i >= 0; i--)
	{
		RGBTRIPLE *rgb = (RGBTRIPLE *) (bitmap + i * width * 3);
		DWORD *lpdwSurf = (DWORD *) surf;

		for (int j = 0; j < width; j++, rgb++)
		{
			DWORD r = rgb->rgbtRed << loRedBit;
			DWORD g = rgb->rgbtGreen << loGrnBit;
			DWORD b = rgb->rgbtBlue << loBluBit;

			lpdwSurf[j] = r | g | b;
		}
		surf += lPitch;
	}
}

int GetPixelFormat(void)
{
	HRESULT r;
	DDPIXELFORMAT format;

	ZeroMemory(&format, sizeof(format));
	format.dwSize = sizeof(format);
	
	r = primary->GetPixelFormat(&format);
	if (r != DD_OK)
	{
		last_error = 0 - (r & 0xFFFF);
		return DMZ_GRAPH_DDRAW_ERROR;
	}

	switch (format.dwRGBBitCount)
	{
	case 15:
		loRedBit = 10;
		loGrnBit = 5;
		loBluBit = 0;
		numRedBits = numGrnBits = numBluBits = 5;
		break;
	case 16:
		loRedBit = 11;
		loGrnBit = 5;
		loBluBit = 0;
		numRedBits = numBluBits = 5;
		numGrnBits = 6;
		break;
	case 24:
	case 32:
		loRedBit = 16;
		loGrnBit = 8;
		loBluBit = 0;
		numRedBits = numGrnBits = numBluBits = 8;
		break;
	default:
		break;
	}
/*
	loRedBit = LowBitPos(format.dwRBitMask);
	WORD hiRedBit = HighBitPos(format.dwRBitMask);
	numRedBits = (WORD) (hiRedBit - loRedBit + 1);

	loGrnBit = LowBitPos(format.dwGBitMask);
	WORD hiGrnBit = HighBitPos(format.dwGBitMask);
	numGrnBits = (WORD) (hiGrnBit - loGrnBit + 1);

	loBluBit = LowBitPos(format.dwBBitMask);
	WORD hiBluBit = HighBitPos(format.dwBBitMask);
	numBluBits = (WORD) (hiBluBit - loBluBit + 1);
*/
	return DMZ_GRAPH_SUCCESS;
}

unsigned char* dmz_graph_load_bitmap(char* filename, int *width, int *height)
{
	BITMAPFILEHEADER tFileHeader;
	BITMAPINFOHEADER tInfoHeader;
	FILE *fp;
	unsigned char *data = NULL;
	int size = 0;

	fp = fopen(filename, "rb");
	if (fp == NULL)
	{
		return NULL;
	}
	
	fseek(fp, 0, SEEK_END);
	size = ftell(fp);
	fseek(fp, 0, SEEK_SET);

	size = size - sizeof (BITMAPFILEHEADER) - sizeof (BITMAPINFOHEADER);
	if (size <= 0)
	{
		fclose(fp);
		return NULL;
	}

	memset(&tFileHeader, 0, sizeof(BITMAPFILEHEADER));

	/* read in the bitmap file header */
	if (fread(&tFileHeader, sizeof(BITMAPFILEHEADER), 1, fp) <= 0)
	{
		fclose(fp);
		return NULL;
	}

	/* check file type */
	if (LOBYTE(tFileHeader.bfType) != 'B' ||
		HIBYTE(tFileHeader.bfType) != 'M')
	{
		fclose(fp);
		return NULL;
	}

	/* read in information header */
	memset(&tInfoHeader, 0, sizeof(BITMAPINFOHEADER));
	if (fread(&tInfoHeader, sizeof(BITMAPINFOHEADER), 1, fp) <= 0)
	{
		fclose (fp);
		return NULL;
	}

	/* check color depth */
	if (tInfoHeader.biBitCount != 24 && tInfoHeader.biBitCount != 8)
	{
		fclose(fp);
		return NULL;
	}

	data = (unsigned char *) malloc(size);

	if (data == NULL)
	{
		fclose(fp);
		return NULL;
	}

	if (fread(data, size, 1, fp) <= 0)
	{
		fclose(fp);
		free(data);
		return NULL;
	}

	if (width != NULL) *width = tInfoHeader.biWidth;
	if (height != NULL) *height = tInfoHeader.biHeight;

	return data;
}

unsigned int dmz_graph_mapRGB(unsigned int red, unsigned int green,
							  unsigned int blue)
{
	float red_div = (float) 256 / (float) pow(2, numRedBits);
	float grn_div = (float) 256 / (float) pow(2, numGrnBits);
	float blu_div = (float) 256 / (float) pow(2, numBluBits);

	float rf = (float) red / red_div;
	float gf = (float) green / grn_div;
	float bf = (float) blue / blu_div;

	DWORD r = ((DWORD) rf << loRedBit);
	DWORD g = ((DWORD) gf << loGrnBit);
	DWORD b = ((DWORD) bf << loBluBit);

	return (r | g | b);
}

int dmz_graph_screenshot(char *filename)
{
	FILE *fp = fopen(filename, "wb");
	BITMAPFILEHEADER bmpFile;
	BITMAPINFOHEADER bmpInfo;
	DDSURFACEDESC surfdesc;
	HRESULT r;

	if (fp == NULL) return DMZ_GRAPH_FILE_ERROR;

	memset(&bmpFile, 0, sizeof(bmpFile));
	memset(&bmpInfo, 0, sizeof(bmpInfo));

	ZeroMemory(&surfdesc, sizeof(surfdesc));
	surfdesc.dwSize = sizeof(surfdesc);

	bmpFile.bfType = 0x4D42;
	bmpFile.bfSize = sizeof(bmpFile);
	bmpFile.bfReserved1 = 0;
	bmpFile.bfReserved2 = 0;
	bmpFile.bfOffBits = sizeof(bmpFile) + sizeof(bmpInfo);

	fwrite(&bmpFile, sizeof(bmpFile), 1, fp);

	bmpInfo.biSize = sizeof(bmpInfo);
	bmpInfo.biWidth = screen_width;;
	bmpInfo.biHeight = screen_height;
	bmpInfo.biPlanes = 1;
	bmpInfo.biBitCount = 24;
	bmpInfo.biCompression = 0;
	bmpInfo.biSizeImage = 0;
	bmpInfo.biXPelsPerMeter = 0;
	bmpInfo.biYPelsPerMeter = 0;
	bmpInfo.biClrImportant = 0;
	bmpInfo.biClrUsed = 0;
	
	fwrite(&bmpInfo, sizeof(bmpInfo), 1, fp);

	while (1)
	{
		r = primary->Lock(NULL, &surfdesc, DDLOCK_READONLY, NULL);

		if (r == DD_OK)
			break;

		else if (r != DDERR_WASSTILLDRAWING)
		{
			last_error = 0 - (r & 0xFFFF);
			return DMZ_GRAPH_DDRAW_ERROR;
		}
	}

	BYTE *lpbSrc = (BYTE *) surfdesc.lpSurface + ((screen_height - 1) * surfdesc.lPitch);

	if (display_depth == 32)
		for (int i = screen_height - 1; i >= 0; i--)
		{
			for (int j = 0; j < screen_width; j++)
			{
				RGBTRIPLE rgb;
				BYTE *colors = lpbSrc + (j << 2);

				rgb.rgbtBlue = colors[0];
				rgb.rgbtGreen = colors[1];
				rgb.rgbtRed = colors[2];
				
				fwrite(&rgb, sizeof(rgb), 1, fp);
			}

			lpbSrc -= surfdesc.lPitch;
		}

	else if (display_depth == 16)
	{
		float red_div = (float) 256 / (float) pow(2, numRedBits);
		float grn_div = (float) 256 / (float) pow(2, numGrnBits);
		float blu_div = (float) 256 / (float) pow(2, numBluBits);

		for (int i = screen_height - 1; i >= 0; i--)
		{
			for (int j = 0; j < screen_width; j++)
			{
				RGBTRIPLE rgb;
				WORD *color = (WORD *) (lpbSrc + (j << 1));

				rgb.rgbtBlue = (BYTE) ((*color & 0x1F) * blu_div);
				rgb.rgbtGreen = (BYTE) (((*color >> numBluBits) & 0x3F) * grn_div);
				rgb.rgbtRed = (BYTE) (((*color >> (numBluBits + numGrnBits)) & 0x1F) * red_div);
				
				fwrite(&rgb, sizeof(rgb), 1, fp);
			}

			lpbSrc -= surfdesc.lPitch;
		}
	}

	primary->Unlock(0);
	fclose(fp);

	return DMZ_GRAPH_SUCCESS;
}

