/*	File:	dmzgraph.cpp
	Author:	Alex Fedosov
	Date:	February 27, 2001
	Description:	alternative graphics library implementation
*/

#include <windows.h>
#include <ddraw.h>

#include <stdio.h>

#include "dmzgraph.h"
#include "dmzgraph_internal.h"

#define DXSAFERELEASE(x) if (x) { x->Release(); x = NULL; }

LPDIRECTDRAW ddraw = NULL;
LPDIRECTDRAWSURFACE primary = NULL, backsurf = NULL;
LPDIRECTDRAWSURFACE images[MAX_IMAGES];
int init_done = 0;
int display_depth = 0, screen_width = 0, screen_height = 0;
int last_error = 0;

int dmz_graph_init(void* hwnd)
{
	HRESULT r;
	int i;

	if (init_done) return DMZ_GRAPH_ALREADY_INIT;

	for (i = 0; i < MAX_IMAGES; i++)
		images[i] = NULL;

	if (hwnd == NULL)
	{
		init_done = 1;
		return DMZ_GRAPH_SUCCESS;
	}

	r = DirectDrawCreate(NULL, &ddraw, NULL);
	if (r != DD_OK)
	{
		last_error = 0 - (r & 0xFFFF);
		return DMZ_GRAPH_INIT_ERROR;
	}

	r = ddraw->SetCooperativeLevel((HWND) hwnd, DDSCL_EXCLUSIVE | DDSCL_FULLSCREEN);
	if (r != DD_OK)
	{
		last_error = 0 - (r & 0xFFFF);
		return DMZ_GRAPH_INIT_ERROR;
	}

	init_done = 1;
	return DMZ_GRAPH_SUCCESS;
}

void dmz_graph_kill(void)
{
	int i;

	if (!init_done) return;

	if (display_depth != 0)
	{

		for (i = 0; i < MAX_IMAGES; i++)
			DXSAFERELEASE(images[i]);

		DXSAFERELEASE(backsurf);
		DXSAFERELEASE(primary);
		DXSAFERELEASE(ddraw);
	}
}

int dmz_graph_set_mode(int width, int height, int depth)
{
	DDSURFACEDESC desc;
	DDSCAPS surfcaps;
	HRESULT r;
	
	if (!init_done) return DMZ_GRAPH_NOT_INIT;

	r = ddraw->SetDisplayMode(width, height, depth);
    if (r != DD_OK)
	{
		last_error = 0 - (r & 0xFFFF);
        return DMZ_GRAPH_MODE_ERROR;
	}

	ZeroMemory(&desc, sizeof(desc));
	desc.dwSize = sizeof(desc);
	desc.dwFlags = DDSD_CAPS | DDSD_BACKBUFFERCOUNT;
	desc.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE | DDSCAPS_FLIP | DDSCAPS_COMPLEX;
	desc.dwBackBufferCount = 1;

	r = ddraw->CreateSurface(&desc, &primary, 0);
	if (r != DD_OK)
	{
		last_error = 0 - (r & 0xFFFF);
		return DMZ_GRAPH_DDRAW_ERROR;
	}

	surfcaps.dwCaps = DDSCAPS_BACKBUFFER;
	r = primary->GetAttachedSurface(&surfcaps, &backsurf);
	if (r != DD_OK)
	{
		last_error = 0 - (r & 0xFFFF);
		return DMZ_GRAPH_DDRAW_ERROR;
	}

	clear_surface(primary, 0, width, height);
	clear_surface(backsurf, 0, width, height);

	display_depth = depth;
	screen_width = width;
	screen_height = height;

	images[0] = backsurf;

	/* get pixel format information */
	GetPixelFormat();

	return DMZ_GRAPH_SUCCESS;
}

int dmz_graph_create_virtual(int width, int height)
{
	DDSURFACEDESC desc;
	HRESULT r;
	int imgID = get_available_image();

	if (!init_done) return DMZ_GRAPH_NOT_INIT;

	if (imgID < 0) return DMZ_GRAPH_TOO_MANY;

	/* hack to allow loading without using DDraw */
	if (display_depth == 0)
	{
		images[imgID] = (LPDIRECTDRAWSURFACE) ~ (int) images[imgID];
		return imgID;
	}

	ZeroMemory(&desc, sizeof(desc));
	desc.dwSize = sizeof(desc);
	desc.dwFlags = DDSD_WIDTH | DDSD_HEIGHT | DDSD_CAPS;
	desc.dwWidth = width;
	desc.dwHeight = height;
	desc.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN | DDSCAPS_SYSTEMMEMORY;

	r = ddraw->CreateSurface(&desc, &images[imgID], 0);

	if (r != DD_OK)
	{
		last_error = 0 - (r & 0xFFFF);
		return DMZ_GRAPH_DDRAW_ERROR;
	}

	clear_surface(images[imgID], 0, width, height);

	return imgID;
}

void dmz_graph_flip(void)
{
	HRESULT r;
	if (!init_done) return;

	r = primary->Flip(NULL, DDFLIP_WAIT);

	if (r == DDERR_SURFACELOST)
	{
		primary->Restore();
		backsurf->Restore();
	}
}

int dmz_text_out(int x, int y, char* textmsg, int textmsglen)
{
	HDC hdc;
	primary->GetDC(&hdc);

	SetBkColor(hdc, RGB(0, 0, 0));
	SetTextColor(hdc, RGB(255, 255, 255));
	SetTextAlign(hdc, TA_LEFT);
	TextOut(hdc, x, y, textmsg, textmsglen);

	primary->ReleaseDC(hdc);

	return DMZ_GRAPH_SUCCESS;
}
